@@ -2,29 +2,33 @@
 SSH utilities for remote client interactions
 """
 import asyncio
+import getpass
 import logging
 import os
 import shlex
+import subprocess
 import sys
 import time
-from contextlib import suppress
-from contextlib import asynccontextmanager
+from contextlib import suppress, contextmanager
 from dataclasses import dataclass
 from pathlib import Path
 from typing import (
     AsyncIterator,
     Callable,
+    ContextManager,
     Dict,
+    List,
     Optional,
     TextIO,
     Tuple,
-    Union, AsyncContextManager, List,
+    Union,
 )
 
 from pytest_mproc import user_output, Settings
 from pytest_mproc.user_output import debug_print, always_print
 
 SUCCESS = 0
+TIMEOUT_SIMPLE_CMD = 15
 
 
 class CommandExecutionFailure(Exception):
@@ -44,7 +48,7 @@ class CommandExecutionFailure(Exception):
 @dataclass
 class SSHClient:
     """
-    SSH Client objct for interacting with a specific client
+    SSH Client class for interacting with a specific client
     """
     host: str
     username: str
@@ -54,9 +58,8 @@ class SSHClient:
 
     def __init__(self, **kwargs):
         self.host = kwargs['host']
-        self.username = kwargs['username']
+        self.username = kwargs.get('username', getpass.getuser())
         self.password = kwargs.get('password')
-        assert self.username
 
     @classmethod
     def set_global_options(cls, *args: str) -> None:
@@ -91,13 +94,13 @@ class SSHClient:
         if env:
             for key, value in env.items():
                 command = f"{key}=\"{value}\" {command}"
-        if user_output.verbose:
+        if user_output.is_verbose:
             env["PTMPROC_VERBOSE"] = '1'
         if cwd is not None:
             command = f"cd {str(cwd)} && {prefix_cmd or ''} {command}"
         command = command.replace('"', '\\\"')
         full_cmd = f"ssh {self.destination} {' '.join(self._global_options)} \"{command}\""
-        debug_print(f"Executing command '{full_cmd}'")
+        debug_print(f"Executing command '%s'", full_cmd)
         if shell:
             result = await asyncio.subprocess.create_subprocess_shell(
                 full_cmd,
@@ -105,7 +108,6 @@ class SSHClient:
                 stderr=stderr,
                 stdin=stdin,
             )
-            assert result
             return result
         else:
             return await asyncio.subprocess.create_subprocess_exec(
@@ -128,11 +130,11 @@ class SSHClient:
             stdout=asyncio.subprocess.PIPE,
             stderr=asyncio.subprocess.STDOUT
         )
-        rc = await asyncio.wait_for(proc.wait(), timeout=5)
+        rc = await asyncio.wait_for(proc.wait(), timeout=TIMEOUT_SIMPLE_CMD)
         if rc != SUCCESS:
             stdout = await proc.stdout.read()
             if not exists_ok or b'File exists' not in stdout:
-                raise CommandExecutionFailure(f"mkdir -p {remote_path}:\n\n{stdout}", rc)
+                raise CommandExecutionFailure(f"mkdir -p {remote_path}:\n\n{stdout.decode('utf-8')}", rc)
 
     async def rmdir(self, remote_path: Path):
         """
@@ -145,7 +147,7 @@ class SSHClient:
             stdout=asyncio.subprocess.DEVNULL,
             stderr=asyncio.subprocess.DEVNULL,
         )
-        rc = await asyncio.wait_for(proc.wait(), timeout=10)
+        rc = await asyncio.wait_for(proc.wait(), timeout=TIMEOUT_SIMPLE_CMD)
         if rc != SUCCESS:
             raise CommandExecutionFailure(f"rm -rf {remote_path}", rc)
 
@@ -161,7 +163,7 @@ class SSHClient:
             stdout=asyncio.subprocess.DEVNULL,
             stderr=asyncio.subprocess.DEVNULL,
         )
-        rc = await asyncio.wait_for(proc.wait(), timeout=10)
+        rc = await asyncio.wait_for(proc.wait(), timeout=TIMEOUT_SIMPLE_CMD)
         if rc != SUCCESS:
             raise CommandExecutionFailure(f"rm -rf {remote_path}", rc)
 
@@ -195,12 +197,14 @@ class SSHClient:
         else:
             rc = await proc.wait()
         if rc != SUCCESS:
-            stdout = await proc.stdout.read()
-            stderr = await proc.stderr.read()
-            raise CommandExecutionFailure(f"Copy from {str(local_path)} to {str(remote_path)} [scp {' '.join(args)}] \n\n{stdout}\n\n {stderr}\n\n",
+            stdout = (await proc.stdout.read()).decode('utf-8')
+            stderr = (await proc.stderr.read()).decode('utf-8')
+            raise CommandExecutionFailure(f"Copy from {str(local_path)} to {str(remote_path)} "
+                                          f"[scp {' '.join(args)}] \n\n{stdout}\n\n {stderr}\n\n",
                                           rc=rc)
 
-    async def pull(self, remote_path: Path, local_path: Path, recursive: bool = True, timeout: Optional[float] = None):
+    async def pull(self, remote_path: Path, local_path: Path,
+                   recursive: bool = True, timeout: Optional[float] = None) -> None:
         """
         Pull file(s) from remote client
 
@@ -209,7 +213,7 @@ class SSHClient:
         :param recursive: specify True if copying a directory
         :param timeout: optional timeout for command to execute
         :raises: TimeoutError if command fails to execute in time
-        :raises: CommendExecutionFailure if command faile to execute on remote host
+        :raises: CommendExecutionFailure if command failed to execute on remote host
         """
         if recursive or local_path.is_dir():
             args = ['-p', '-r',  f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
@@ -217,6 +221,7 @@ class SSHClient:
             args = ['-p', f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
         args = list(self._global_options) + args
         cmd = f"scp {' '.join(args)}"
+        debug_print(f"Executing '{cmd}'")
         proc = await asyncio.create_subprocess_shell(
             cmd,
             stdout=asyncio.subprocess.PIPE,
@@ -227,12 +232,46 @@ class SSHClient:
         else:
             rc = await proc.wait()
         if rc != 0:
-            stdout = await proc.stdout.read()
-            msg = f"\n!!! Copy from  {self.destination}:{str(remote_path)} to {str(local_path.absolute())} [{cmd}]\n"\
-                  f"\n{stdout.decode('utf-8')}\n\n"
+            stdout = (await proc.stdout.read()).decode('utf-8')
+            msg = f"\n!!! Copy failed from  {self.destination}:{str(remote_path)} to "\
+                  f"{str(local_path.absolute())} [{cmd}]\n"\
+                  f"\n{stdout}\n\n"
             always_print(msg)
             raise CommandExecutionFailure(msg, rc)
 
+    def pull_sync(self, remote_path: Path, local_path: Path,
+                  recursive: bool = True, timeout: Optional[float] = None) -> None:
+        """
+        Pull file(s) from remote client
+
+        :param remote_path: file or directory to pull (recursively) from remote client
+        :param local_path: path to copy to local host
+        :param recursive: specify True if copying a directory
+        :param timeout: optional timeout for command to execute
+        :raises: TimeoutError if command fails to execute in time
+        :raises: CommendExecutionFailure if command failed to execute on remote host
+        """
+        if recursive or local_path.is_dir():
+            args = ['-p', '-r',  f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
+        else:
+            args = ['-p', f"{self.destination}:{str(remote_path)}", str(local_path.absolute())]
+        args = list(self._global_options) + args
+        cmd = f"scp {' '.join(args)}"
+        debug_print(f"Executing '{cmd}'")
+        proc = subprocess.run(
+            cmd,
+            stdout=asyncio.subprocess.PIPE,
+            stderr=asyncio.subprocess.STDOUT,
+            shell=True,
+            timeout=timeout
+        )
+        if proc.returncode != 0:
+            stdout = proc.stdout.decode('utf-8')
+            msg = f"\n!!! Copy from  {self.destination}:{str(remote_path)} to {str(local_path.absolute())} [{cmd}]\n"\
+                  f"\n{stdout}\n\n"
+            always_print(msg)
+            raise CommandExecutionFailure(msg, proc.returncode)
+
     async def get_remote_platform_info(self):
         """
         :return:  Tuple of system and machine platform attributes of remote client
@@ -242,7 +281,7 @@ class SSHClient:
             stdout=asyncio.subprocess.PIPE,
             stderr=asyncio.subprocess.DEVNULL,
         )
-        stdout, _ = await asyncio.wait_for(proc.communicate(), timeout=5)
+        stdout, _ = await asyncio.wait_for(proc.communicate(), timeout=TIMEOUT_SIMPLE_CMD)
         if proc.returncode != 0:
             raise CommandExecutionFailure("Getting platform info through uname", proc.returncode)
         system, machine = stdout.strip().splitlines()
@@ -253,54 +292,46 @@ class SSHClient:
 
     async def install_packages(
             self,  *pkgs,
-            venv: Path,
-            cwd: Optional[Path] = None,
+            remote_executable: str,
             stdout: Optional[Union[int, TextIO]] = None,
             stderr: Optional[Union[int, TextIO]] = sys.stderr):
-        remote_py_executable = venv / 'bin' / 'python3'
-        cmd = f"{str(remote_py_executable)} -m pip install --upgrade {' '.join(pkgs)}"
+        cmd = f"{remote_executable} -m pip install --upgrade {' '.join(pkgs)}"
         proc = await self._remote_execute(
             cmd,
             stdout=stdout,
             stderr=stderr,
-            cwd=str(cwd) if cwd is not None else None
         )
         stdout, _ = await proc.communicate()
         assert proc.returncode is not None
         if proc.returncode != 0:
-            os.write(sys.stderr.fileno(), f"Command {cmd} failed to install on remote".encode('utf-8'))
+            always_print(f"Command {cmd} failed to install on remote", as_error=True)
             raise CommandExecutionFailure(cmd, proc.returncode)
 
     async def install(self,
                       venv: Path,
-                      remote_root: Path,
-                      requirements_path: str,
-                      cwd: Optional[Path] = None,
+                      full_requirements_path: Path,
                       stdout: Optional[Union[int, TextIO]] = None,
                       stderr: Optional[Union[int, TextIO]] = sys.stderr):
         """
         install Python requirements  on remote client
 
-        :param cwd: directory to run from on remote host
         :param venv: path to python virtual environment on remote host
-        :param remote_root: root directory on remote client
-        :param requirements_path: path, relative to remote_root, where requirments file is to be found
+        :param full_requirements_path: path, relative to remote_root, where requirements file is to be found
         :param stdout: as per subprocess.Popen
         :param stderr: as per subprocess.Popen
-        :raises: CommandExecutionFailure if install failes on remote client
+        :raises: CommandExecutionFailure if install failed on remote client
         """
         remote_py_executable = venv / 'bin' / 'python3'
-        cmd = f"{str(remote_py_executable)} -m pip install --upgrade -r {str(remote_root / requirements_path)}"
+        cmd = f"{str(remote_py_executable)} -m pip install --upgrade -r {str(full_requirements_path)}"
         proc = await self._remote_execute(
             cmd,
             stdout=stdout,
             stderr=stderr,
-            cwd=str(cwd) if cwd is not None else None
         )
         stdout, _ = await proc.communicate()
         assert proc.returncode is not None
         if proc.returncode != 0:
-            os.write(sys.stderr.fileno(), f"Command {cmd} failed to install on remote".encode('utf-8'))
+            always_print(f"Command {cmd} failed to install on remote", as_error=True)
             raise CommandExecutionFailure(cmd, proc.returncode)
 
     async def monitor_remote_execution(self, command, *args,
@@ -317,7 +348,7 @@ class SSHClient:
         :param args: args to the command
         :param timeout: optional timeout for command
         :param cwd: directory on remote client in which to execute
-        :param success_code: None if not to check return code, otherwise lambda taking return code an returning
+        :param success_code: None if not to check return code, otherwise lambda taking return code and returning
            True is success or False otherwise (defaults to 0 being success code)
         :param stdin: stream to use for stdin (as in subprocess.Popen)
         :param env: optional dict of environment vars to apply
@@ -328,23 +359,23 @@ class SSHClient:
         """
         start_time = time.monotonic()
 
-        def getlines(data: str) -> Tuple[str, str]:
+        def getlines(data_: str) -> Tuple[str, str]:
             """
             Split data into lines, but treat any last remnant as a buffer for client to combine with
             a next batch of data before declaring a full line
 
-            :param data: string to process
+            :param data_: string to process
             :return: a tuple of lines contained in a string and a buffer of data leftover at end
             """
-            if '\n' in data:
-                *lines, buffer = data.splitlines()
-                if data.endswith('\n'):
-                    lines.append(buffer)
-                    return lines, ""
+            if '\n' in data_:
+                *lines_, buffer = data_.splitlines()
+                if data_.endswith('\n'):
+                    lines_.append(buffer)
+                    return lines_, ""
                 else:
-                    return lines, buffer
+                    return lines_, buffer
             else:
-                return "", data
+                return "", data_
         proc = await self.launch_remote_command(command, *args,
                                                 stdin=stdin,
                                                 stdout=asyncio.subprocess.PIPE,
@@ -363,8 +394,8 @@ class SSHClient:
         if timeout is not None and time_left <= 0.0:
             with suppress(OSError):
                 proc.terminate()
-            with suppress(TimeoutError):
-                await asyncio.wait_for(proc.wait(), timeout=5)
+            with suppress(TimeoutError, KeyboardInterrupt):
+                await asyncio.wait_for(proc.wait(), timeout=TIMEOUT_SIMPLE_CMD)
             if proc.returncode is None:
                 with suppress(OSError):
                     proc.kill()
@@ -376,34 +407,34 @@ class SSHClient:
             logging.error(f"process '{command} {' '.join(args)}' failed with code {proc.returncode}")
             raise CommandExecutionFailure(command, proc.returncode)
 
-    @asynccontextmanager
-    async def mkdtemp(self, tmp_root: Optional[Path] = None) -> Path:
+    @contextmanager
+    def mkdtemp(self, tmp_root: Optional[Path] = None) -> Path:
         """
         :return: temporary directory created on remote client
         """
         if tmp_root is None:
-            proc = await self._remote_execute(
+            # noinspection SpellCheckingInspection
+            proc = self._remote_execute_sync(
                 f"mktemp -d --tmpdir ptmproc-tmp.XXXXXXXXXX",
                 stdout=asyncio.subprocess.PIPE,
                 stderr=asyncio.subprocess.PIPE,
                 shell=True
             )
         else:
-            proc = await self._remote_execute(
+            # noinspection SpellCheckingInspection
+            proc = self._remote_execute_sync(
                 f"mktemp -d  {str(tmp_root / 'ptmproc-tmp.XXXXXXXXXX')}",
                 stdout=asyncio.subprocess.PIPE,
                 stderr=asyncio.subprocess.PIPE,
                 shell=True
             )
-        await asyncio.wait_for(proc.wait(), timeout=5)
+        stdout, stderr = proc.communicate(timeout=TIMEOUT_SIMPLE_CMD)
         if proc.returncode != 0:
-            stdout = await proc.stdout.read()
-            stderr = await proc.stderr.read()
             raise SystemError(f"Failed to create tmp dir on remote client: {stdout}\n  {stderr}")
         # noinspection SpellCheckingInspection
-        tmpdir = (await proc.stdout.read()).decode('utf-8').strip()
+        tmpdir = stdout.decode('utf-8').strip()
         yield Path(tmpdir)
-        await self.rmdir(Path(tmpdir))
+        self.rmdir_sync(Path(tmpdir))
 
     async def launch_remote_command(self,  command, *args,
                                     stdin: Union[TextIO, int],
@@ -413,7 +444,7 @@ class SSHClient:
                                     env: Optional[Dict[str, str]] = None,
                                     prefix_cmd: Optional[str] = None,
                                     auth_key: Optional[bytes] = None,
-                                    shell:bool = True) -> asyncio.subprocess.Process:
+                                    shell: bool = True) -> asyncio.subprocess.Process:
         """
         execute command on remote host
 
@@ -426,11 +457,12 @@ class SSHClient:
         :param stderr: stream to use for stderr (as in subprocess.Popen)
         :param env: optional dictionary of environment variables for command
         :param auth_key: authentication key for multiprocessing
+        :param shell: whether to run in shell or not
         :return: Process created (completed process if timeout is specified)
         :raises: TimeoutError if command does not execute in time (if timeout is specified)
         """
-        args = [shlex.quote(arg) if '*' not in arg else arg for arg in args]
-        full_command = command + ' ' + ' '.join(args)
+        args = [shlex.quote(str(arg)) if '*' not in arg else arg for arg in args]
+        full_command = str(command) + ' ' + ' '.join(args)
         result = await self._remote_execute(
             full_command,
             stdin=stdin,
@@ -448,7 +480,7 @@ class SSHClient:
     async def execute_remote_cmd(self,  command, *args,
                                  stdout: Union[int, TextIO] = sys.stdout,
                                  stderr: Union[int, TextIO] = sys.stderr,
-                                 stdin: Optional[TextIO] = None,
+                                 stdin: Optional[Union[TextIO, int]] = None,
                                  timeout: Optional[float] = None,
                                  cwd: Optional[Path] = None,
                                  prefix_cmd: Optional[str] = None,
@@ -470,6 +502,7 @@ class SSHClient:
         :param stderr: stream to use for stderr (as in subprocess.Popen)
         :param env: optional dictionary of environment variables for command
         :param auth_key: authentication key for multiprocessing
+        :param shell: whether to run in shell or not
         :return: Process created (completed process if timeout is specified)
         :raises: TimeoutError if command does not execute in time (if timeout is specified)
         """
@@ -490,36 +523,297 @@ class SSHClient:
         await asyncio.wait_for(proc.wait(), timeout=timeout)
         return proc
 
+    def signal_sync(self, pgids: List[int], sig_number: int) -> None:
+        """
+        Kill processes with assigned pids on remote host.
 
-@asynccontextmanager
-async def remote_root_context(project_name: str, ssh_client: SSHClient, remote_root: Optional[Path]) -> \
-        AsyncContextManager[Tuple[Path, Path]]:
+        :raises CommandExecutionFailure: if it fails to kill, with exception if process is no longer active;  in that
+           case this function returns normally
+        """
+        pid_texts = [str(-pgid) for pgid in pgids]
+        try:
+            command = f"kill -{sig_number} {' '.join(pid_texts)}"
+            completed = subprocess.run(
+                ["ssh", self.destination, *self._global_options, command],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.STDOUT,
+                timeout=TIMEOUT_SIMPLE_CMD
+            )
+            if completed.returncode != 0 and b'No such process' not in completed.stdout:
+                always_print(f"Failed to kill one or more worker pids {' '.join(pid_texts)} "
+                             f"on {self.destination}: {completed.stdout.decode('utf-8')}", as_error=True)
+        except TimeoutError:
+            always_print("Timeout trying to kill remote worker processes")
+
+    async def port_forward(self, local_port: int, remote_port: int):
+        """
+        Port forward from local machine to remote port
+        :param local_port: port on local machine to forward connection requests
+        :param remote_port: port on remote machine to accept (forwarded) connections requests
+        """
+        debug_print(f"Executing command 'ssh -N -L {local_port}:127.0.0.1:{remote_port} {self.destination}'")
+        proc = await asyncio.subprocess.create_subprocess_exec(
+            "ssh", "-N", "-L", f"{local_port}:127.0.0.1:{remote_port}", f"{self.destination}",
+            stdout=asyncio.subprocess.PIPE,
+            stderr=asyncio.subprocess.STDOUT
+        )
+        await asyncio.sleep(0.5)
+        if proc.returncode is not None:
+            stdout = (await proc.stdout.read()).decode('utf-8')
+            raise CommandExecutionFailure(f"Failed to port forward: {stdout}", rc=proc.returncode)
+        return proc
+
+    async def reverse_port_forward(self, local_port: int, remote_port: int):
+        """
+        Reverse port forward from remote port to local machine
+        :param local_port: port on local machine to accept (reverse forwarded) connection requests
+        :param remote_port: port on remote machine to  forward connections requests
+        """
+        debug_print(f"Executing command 'ssh -N -R {remote_port}:127.0.0.1:{local_port} {self.destination}'")
+        proc = await asyncio.subprocess.create_subprocess_exec(
+            "ssh", "-N", "-R", f"{remote_port}:127.0.0.1:{local_port}", f"{self.destination}",
+            stdout=asyncio.subprocess.PIPE,
+            stderr=asyncio.subprocess.STDOUT
+        )
+        await asyncio.sleep(0.5)
+        if proc.returncode is not None:
+            stdout = (await proc.stdout.read()).decode('utf-8')
+            raise CommandExecutionFailure(f"Failed to port forward: {stdout}", rc=proc.returncode)
+        return proc
+
+    async def find_free_port(self) -> int:
+        """
+        find free port on remote host
+        """
+        # noinspection SpellCheckingInspection
+        proc = await self._remote_execute(
+            "python3 -c \"import socket; s = socket.socket(); s.bind(('', 0));print(s.getsockname()[1]);s.close()\"",
+            stdout=asyncio.subprocess.PIPE,
+            stderr=sys.stderr,
+            shell=True
+        )
+        text = (await proc.stdout.read()).decode('utf-8')
+        return int(text.strip())
+
+    def _remote_execute_sync(
+        self, command: str,
+        stdout, stderr, stdin=None,
+        env=None,
+        cwd=None,
+        prefix_cmd: Optional[str] = None,
+        auth_key: Optional[bytes] = None,
+        shell: bool = True
+    ) -> subprocess.Popen:
+        """
+        See equivalent async method docs
+        """
+        env = env or {}
+        if auth_key is not None:
+            env["AUTH_TOKEN_STDIN"] = '1'
+        if env:
+            for key, value in env.items():
+                command = f"{key}=\"{value}\" {command}"
+        if user_output.is_verbose:
+            env["PTMPROC_VERBOSE"] = '1'
+        if cwd is not None:
+            command = f"cd {str(cwd)} && {prefix_cmd or ''} {command}"
+        command = command.replace('"', '\\\"')
+        full_cmd = f"ssh {self.destination} {' '.join(self._global_options)} \"{command}\""
+        debug_print(f"Executing command '%s'", full_cmd)
+        if shell:
+            result = subprocess.Popen(
+                full_cmd,
+                stdout=stdout,
+                stderr=stderr,
+                stdin=stdin,
+                shell=True,
+            )
+            assert result
+            return result
+        else:
+            return subprocess.Popen(
+                ['ssh', self.destination, *self._global_options, f"\"command\""],
+                stdout=stdout,
+                stderr=stderr,
+                stdin=stdin,
+            )
+
+    def mkdir_sync(self, remote_path: Path, exists_ok=True):
+        """
+        See mkdir method docs
+        """
+        options = "-p" if exists_ok is True else ""
+        proc = self._remote_execute_sync(
+            f"mkdir {options} {str(remote_path)}",
+            stdout=asyncio.subprocess.PIPE,
+            stderr=asyncio.subprocess.STDOUT
+        )
+        rc = proc.wait(timeout=TIMEOUT_SIMPLE_CMD)
+        if rc != SUCCESS:
+            stdout = proc.stdout.read()
+            if not exists_ok or b'File exists' not in stdout:
+                raise CommandExecutionFailure(f"mkdir -p {remote_path}:\n\n{stdout.decode('utf-8')}", rc)
+
+    def rmdir_sync(self, remote_path: Path):
+        """
+        See equivalent async method docs
+        """
+        proc = self._remote_execute_sync(
+            f"rm -r {remote_path}",
+            stdout=asyncio.subprocess.DEVNULL,
+            stderr=asyncio.subprocess.DEVNULL,
+        )
+        rc = proc.wait(timeout=TIMEOUT_SIMPLE_CMD)
+        if rc != SUCCESS:
+            raise CommandExecutionFailure(f"rm -rf {remote_path}", rc)
+
+    def launch_remote_command_sync(
+        self,
+        command,
+        *args,
+        stdin: Union[TextIO, int],
+        stdout: Union[TextIO, int],
+        stderr: Union[TextIO, int],
+        cwd: Optional[Path] = None,
+        env: Optional[Dict[str, str]] = None,
+        prefix_cmd: Optional[str] = None,
+        auth_key: Optional[bytes] = None,
+        shell: bool = True
+    ) -> subprocess.Popen:
+        """
+        execute command on remote host
+
+        See equivalent async method docs
+        """
+        args = [shlex.quote(str(arg)) if '*' not in arg else arg for arg in args]
+        full_command = str(command) + ' ' + ' '.join(args)
+        return self._remote_execute_sync(
+            full_command,
+            stdin=stdin,
+            stdout=stdout,
+            stderr=stderr,
+            env=env,
+            cwd=cwd,
+            prefix_cmd=prefix_cmd,
+            auth_key=auth_key,
+            shell=shell
+        )
+
+    def execute_remote_cmd_sync(
+            self,  command, *args,
+            stdout: Union[int, TextIO] = sys.stdout,
+            stderr: Union[int, TextIO] = sys.stderr,
+            stdin: Optional[Union[TextIO, int]] = None,
+            timeout: Optional[float] = None,
+            cwd: Optional[Path] = None,
+            prefix_cmd: Optional[str] = None,
+            env: Optional[Dict[str, str]] = None,
+            auth_key: Optional[bytes] = None,
+            shell: bool = True,
+    ) -> subprocess.Popen:
+        """
+        execute command on remote host
+
+        See equivalent async method docs
+        """
+        proc = self.launch_remote_command_sync(
+            command, *args,
+            stdin=asyncio.subprocess.PIPE if auth_key is not None else stdin,
+            stdout=stdout,
+            stderr=stderr,
+            cwd=cwd,
+            auth_key=auth_key,
+            prefix_cmd=prefix_cmd,
+            shell=shell,
+            env=env)
+        if auth_key is not None:
+            import binascii
+            auth_key = binascii.b2a_hex(auth_key)
+            proc.stdin.write(auth_key + b'\n')
+            proc.stdin.close()
+        proc.wait(timeout=timeout)
+        return proc
+
+    async def ping(self, timeout: float) -> bool:
+        try:
+            await self.execute_remote_cmd(
+                "echo", "PONG",
+                stdout=asyncio.subprocess.DEVNULL,
+                stderr=asyncio.subprocess.DEVNULL,
+                timeout=timeout,
+                shell=True,
+            )
+            return True
+        except (TimeoutError, asyncio.TimeoutError):
+            return False
+
+    async def unzip(self, remote_zip_location: Union[Path, str], remote_target_dir: Path,
+                    timeout: Optional[float] = None):
+        proc = await self._remote_execute(
+            f"cd {str(remote_target_dir)} && unzip -o {str(remote_zip_location)}",
+            stdout=sys.stdout if user_output.is_verbose else asyncio.subprocess.DEVNULL,
+            stderr=sys.stderr
+        )
+        await asyncio.wait_for(proc.wait(), timeout=timeout)
+        if proc.returncode != 0:
+            text = ('\n' + (await proc.stdout.read()).decode('utf-8')) if proc.stdout else ""
+            raise CommandExecutionFailure(f"Failed to unzip tests on remote client {text}", proc.returncode)
+
+    async def remote_file_exists(self, remote_path: Path):
+        proc = await self.execute_remote_cmd(
+            'ls', str(remote_path),
+            stderr=asyncio.subprocess.DEVNULL,
+            stdout=asyncio.subprocess.DEVNULL,
+        )
+        return proc.returncode == 0
+
+    async def create_remote_venv(self, remote_py_executable: Union[Path, str], remote_venv_path: Path):
+        proc = await self.execute_remote_cmd(
+            str(remote_py_executable), '-m', 'venv', str(remote_venv_path),
+            stdout=sys.stdout if user_output.is_verbose else asyncio.subprocess.DEVNULL,
+            stderr=asyncio.subprocess.PIPE
+        )
+        if proc.returncode != 0:
+            text = ('\n' + (await proc.stderr.read()).decode('utf-8')) if proc.stderr else ""
+            raise CommandExecutionFailure(f"Failed to unzip tests on remote client {text}", proc.returncode)
+
+    async def touch(self, remote_file_path: Union[Path, str]):
+        proc = await self.execute_remote_cmd(
+            f'touch {str(remote_file_path)}',
+            stdout=sys.stdout if user_output.is_verbose else asyncio.subprocess.DEVNULL,
+            stderr=asyncio.subprocess.PIPE,
+            shell=True
+        )
+        if proc.returncode != 0:
+            text = ('\n' + (await proc.stderr.read()).decode('utf-8')) if proc.stderr else ""
+            raise CommandExecutionFailure(f"Failed to unzip tests on remote client {text}", proc.returncode)
+
+
+@contextmanager
+def remote_root_context(project_name: str, ssh_client: SSHClient, remote_root: Optional[Path]) -> \
+        ContextManager[Tuple[Path, Path]]:
     """
     Provide a context representing a directory on the remote host from which to run.  If no explicit
     path is provided as a remote root, a temporary directory is created on remote host and removed
     when exiting this context.  (Otherwise context manager has no real effect)
 
-    :param project_name: name of project, used in createing a cache dir if available to remote system
+    :param project_name: name of project, used in creating a cache dir if available to remote system
     :param ssh_client: ssh client to use in creating a temp dir if necessary
-    :param remote_root: an explicit location on the remote host (will not be remove on exit of context)
+    :param remote_root: an explicit location on the remote host (will not be removed on exit of context)
     :return: the remote root created (or reflects the remote root provided explicitly)
     """
-    remote_root = remote_root or Settings.tmp_root
-    if not Settings.cache_dir:
-        proc = await ssh_client.execute_remote_cmd('echo \${HOME}', stdout=asyncio.subprocess.PIPE, shell=True)
-        stdout_text = (await proc.stdout.read()).strip().decode('utf-8')
-        if proc.returncode != 0 or not stdout_text:
+    if Settings.cache_dir is None:
+        proc = ssh_client.execute_remote_cmd_sync('pwd', stdout=asyncio.subprocess.PIPE, shell=True)
+        stdout_text = proc.stdout.read().strip().decode('utf-8')
+        if (proc.returncode != 0 or not stdout_text) and remote_root is not None:
             cache_path = remote_root
         else:
             cache_path = Path(stdout_text) / '.ptmproc' / 'cache'
     else:
-        cache_path = Settings.cache_dir or remote_root
+        cache_path = Settings.cache_dir
     venv_root = cache_path / project_name
-    await ssh_client.mkdir(venv_root, exists_ok=True)
+    ssh_client.mkdir_sync(venv_root, exists_ok=True)
     debug_print(f"Using {cache_path / project_name} to cache python venv")
-    if remote_root:
-        yield remote_root, venv_root or remote_root
-    else:
-        async with ssh_client.mkdtemp(Settings.tmp_root) as root:
-            debug_print(f"Using dir {venv_root} to set up python venv and {root} as temp staging dir")
-            yield root, venv_root or root
+    with ssh_client.mkdtemp(remote_root or Settings.tmp_root) as root:
+        always_print(f"Using dir {venv_root} to set up python venv and {root} as temp staging dir")
+        yield root, venv_root or root
